unit u_calc;

// Purpose: Calculations on Hall Data

{$mode objfpc}{$H+}

interface

uses Classes, SysUtils;

var
  EllipseShortAxis_mm,
  PrecessionAngle_Degrees: real;

procedure ProcessHallData;

implementation

uses u_main, u_messages, u_logging, u_compass, u_numstrings, math, graphics;

var
  PosNS, PosEW, PrevPosNS, PrevPosEW,
  DistanceSquared, PrevDistanceSquared,
  avFarestPosNS, avFarestPosEW,
  avShortAxisNS, avShortAxisEW,
  EllipseShortAxis,
  EllipseRatio_Percent,
  EllipseCalibrationFactor, EllipseLongAxis,
  PrecessionAngle_Radians: real;
  HaveFirst, FoundFarest: boolean;
  CompassScale, CalibrationScale: integer;

procedure ProcessHallData; // executed at each message arrival
var
  RotX, RotY,
  avFactor: real;
begin
avFactor:= Fival (frmMain.TbxAverageFactor_Percent.text) / 100;
if CalibrateHallSensors then
   begin
     // For adjusting the offsets, while Bob is still, we use the
     // enlargement scale for fine adjustment.
     if Enlarge then CalibrationScale:= 10 else CalibrationScale:= 2;
     // We will adjust the amplitude of the individual Hall Channels the
     // same value when the bob is at a known distance from the center
     if not Persistence then Compass.Init (FrmMain, 520, 3, 330);
     Compass.PlotFilledBox ((AdcHallEast - 128) *       CalibrationScale, 0, 3, clYellow);  // color sunrise
     Compass.PlotFilledBox ((AdcHallWest - 128) *      -CalibrationScale, 0, 3, clRed);     // color sunset
     Compass.PlotFilledBox ( 0, (AdcHallNorth - 128) *  CalibrationScale,    3, clFuchsia); // color northern light
     Compass.PlotFilledBox ( 0, (AdcHallSouth - 128) * -CalibrationScale,    3, clAqua);    // color ice
   end
   else
   begin
     // Calculate normalized positions from Hall sensors
     // Normalized means they are between -1 and +1
     // The > 0 test is to prevent a possible div by zero crash
     if (AdcHallNorth + AdcHallSouth) > 0 then
       PosNS:= (real(AdcHallNorth - AdcHallSouth) / real(AdcHallNorth + AdcHallSouth));
     if (AdcHallEast + AdcHallWest) > 0 then
       PosEW:= (real(AdcHallEast - AdcHallWest) / real(AdcHallEast + AdcHallWest));
     // Show normalized positions
     FrmMain.TbxPosNS.text:= format ('%5.3f', [PosNS]);
     Frmmain.TbxPosEW.text:= format ('%5.3f', [PosEW]);
     // Plot on Compass display
     CompassScale:= fIval (FrmMain.TbxCompassScale.text);
     if not Persistence then
     begin // remove previous blip
       Compass.PlotPoint (round (PrevPosEW * CompassScale), round (PrevPosNS * CompassScale), 1, clBlack);
       Compass.ReDraw; // redraw grid
     end;
     if HalfSwing then // show active halfswing in aqua color
       Compass.PlotPoint (round (PosEW * CompassScale), round (PosNS * CompassScale), 1, clAqua) // new blip
     else
       Compass.PlotPoint (round (PosEW * CompassScale), round (PosNS * CompassScale), 1, clWhite);
     // save previous positions for removal in plot and for precession angle calculation.
     PrevPosNS:= PosNS;
     PrevPosEW:= PosEW;
     // Plot Rimpasses for diagnosis
     if PlotRim_1 then if SeenRim_1 then
       Compass.DrawCircle (round (PosEW * CompassScale), round (PosNS * CompassScale), 2, clRed);
     if PlotRim_2 then if SeenRim_2 then
       Compass.DrawCircle (round (PosEW * CompassScale), round (PosNS * CompassScale), 2, clRed);

     if HalfSwing then // We do further processing only for one HalfSwing
     begin
       // FrmMain.memo.lines.add ('HalfSwing');
       if SeenCenter then // prepare for searches
       begin
         //FrmMain.memo.lines.add ('SeenCenter');
         PrevDistanceSquared:= 0; // to start a new search
         FoundFarest:= false;     // for finding the precession angle
         HaveFirst:= false;       // for finding the short axis of the ellipse
       end;

       // To find the precession angle we search for the farest distance from the center
       if not FoundFarest then
       begin
         DistanceSquared:= sqr (PosEW) + sqr (PosNS); // we dont need the squareroot
         if DistanceSquared >= PrevDistanceSquared then // track increasing distance
           PrevDistanceSquared:= DistanceSquared
         else // we are on the way back, so the previous position was the farest away
         begin
           FoundFarest:= true;
           // Enter the farest coordinates into the leaking bucket averagers
           avFarestPosNS:= avFarestPosNS * (1-avFactor) + PrevPosNS * avFactor;
           avFarestPosEW:= avFarestPosEW * (1-avFactor) + PrevPosEW * avFactor;
           // Plot the farest position averaged
           Compass.PlotFilledBox (round (avFarestPosEW * CompassScale),
                                  round (avFarestPosNS * CompassScale), 2, clAqua); // new blip
           FrmMain.TbxFarestNS.text:= format ('%5.3f', [avFarestPosNS]);
           FrmMain.TbxFarestEW.text:= format ('%5.3f', [avFarestPosEW]);
           // calculate the angle
           PrecessionAngle_Radians:= arctan2 (avFarestPosNS, avFarestPosEW);
           PrecessionAngle_Degrees:= PrecessionAngle_Radians * 180 / pi();
           FrmMain.TbxPrecessionAngle.text:= format ('%5.1f', [PrecessionAngle_Degrees]);
           Compass.PlotAngleMarker (PrecessionAngle_Degrees, clRed);
           // To have calibrated values for the Ellipse dimensions we take the longest axis in Hall-units:
           EllipselongAxis:= sqrt (PrevDistanceSquared);
           FrmMain.TbxEllipseLongAxis.text:= format ('%5.3f', [EllipseLongAxis]);
           // But we know that this is equal to the Setpoint, if Amplitude Control is working properly
           if EllipseLongAxis > 0.00000001 then // prevent div by nearly zero
             EllipseCalibrationFactor:= SetPoint_Amplitude_mm / EllipseLongAxis;
         end;
       end;

       // We calculate the Ellipse Short Axis from the first sample after a Center Pass
       if not HaveFirst then
       begin
         HaveFirst:= true;
         // and put it into the leaking bucket averagers.
         avShortAxisNS:= avShortAxisNS * (1-avFactor) + PosNS * avFactor;
         avShortAxisEW:= avShortAxisEW * (1-avFactor) + PosEW * avFactor;
         // The short axis is the distance between the center and this sample
         EllipseShortAxis:= sqrt (sqr (avShortAxisNS) + sqr (avShortAxisEW)); // Hall units
         Compass.PlotFilledBox (round (PrevPosEW * CompassScale), round (PrevPosNS * CompassScale), 2, clLime); // new blip
         // To find the polarity of the short axis, we rotate the sample over -PrecessionAngle.
         // Otherwise said: the ellipse is now laying horizontally (if all points were rotated).
         Rotate (avShortAxisEW, avShortAxisNS, -PrecessionAngle_Radians, RotX, RotY);
         Compass.PlotPoint (round (RotX * CompassScale), round (RotY * CompassScale), 1, clRed);
         if RotY > 0 then EllipseShortAxis *= -1; // Positive rotation is CW
         EllipseRatio_Percent:= 100.0 * EllipseShortAxis / EllipseLongAxis;
         FrmMain.TbxEllipseRatio_Percent.text:= format ('%5.2f', [EllipseRatio_Percent]);
         // Calibrate the short axis:
         EllipseShortAxis_mm:= EllipseShortAxis * EllipseCalibrationFactor;
         FrmMain.TbxEllipseShortAxis.text:= format ('%6.2f', [EllipseShortAxis_mm]);
       end;
       if SeenCenter then LogData;
     end; // if HalfSwing
   end;
end;

end. // End of Unit

